"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const serum_1 = require("@project-serum/serum");
const fixednum_1 = require("./fixednum");
const layout_1 = require("./layout");
const utils_1 = require("./utils");
const MangoGroup_1 = require("./MangoGroup");
class MangoAccount {
    constructor(publicKey, decoded) {
        this.publicKey = publicKey;
        this.spotOpenOrdersAccounts = new Array(layout_1.MAX_PAIRS).fill(undefined);
        Object.assign(this, decoded);
    }
    reload(connection) {
        return __awaiter(this, void 0, void 0, function* () {
            const acc = yield connection.getAccountInfo(this.publicKey);
            Object.assign(this, layout_1.MangoAccountLayout.decode(acc === null || acc === void 0 ? void 0 : acc.data));
            return this;
        });
    }
    // TODO: use getMultipleAccounts instead
    loadOpenOrders(connection, serumDexPk) {
        return __awaiter(this, void 0, void 0, function* () {
            const promises = [];
            for (let i = 0; i < this.spotOpenOrders.length; i++) {
                if (this.spotOpenOrders[i].equals(utils_1.zeroKey)) {
                    promises.push(utils_1.promiseUndef());
                }
                else {
                    promises.push(serum_1.OpenOrders.load(connection, this.spotOpenOrders[i], serumDexPk));
                }
            }
            this.spotOpenOrdersAccounts = yield Promise.all(promises);
            return this.spotOpenOrdersAccounts;
        });
    }
    getNativeDeposit(rootBank, tokenIndex) {
        return rootBank.depositIndex.mul(this.deposits[tokenIndex]);
    }
    getNativeBorrow(rootBank, tokenIndex) {
        return rootBank.borrowIndex.mul(this.borrows[tokenIndex]);
    }
    getUiDeposit(rootBank, mangoGroup, tokenIndex) {
        return utils_1.nativeI80F48ToUi(this.getNativeDeposit(rootBank, tokenIndex), mangoGroup.tokens[tokenIndex].decimals);
    }
    getUiBorrow(rootBank, mangoGroup, tokenIndex) {
        return utils_1.nativeI80F48ToUi(this.getNativeBorrow(rootBank, tokenIndex), mangoGroup.tokens[tokenIndex].decimals);
    }
    getSpotVal(mangoGroup, mangoCache, index, assetWeight) {
        let assetsVal = fixednum_1.ZERO_I80F48;
        const price = mangoGroup.getPrice(index, mangoCache);
        const depositVal = this.getUiDeposit(mangoCache.rootBankCache[index], mangoGroup, index)
            .mul(price)
            .mul(assetWeight);
        assetsVal = assetsVal.add(depositVal);
        const openOrdersAccount = this.spotOpenOrdersAccounts[index];
        if (openOrdersAccount !== undefined) {
            assetsVal = assetsVal.add(fixednum_1.I80F48.fromNumber(utils_1.nativeToUi(openOrdersAccount.baseTokenTotal.toNumber(), mangoGroup.tokens[index].decimals))
                .mul(price)
                .mul(assetWeight));
            assetsVal = assetsVal.add(fixednum_1.I80F48.fromNumber(utils_1.nativeToUi(openOrdersAccount.quoteTokenTotal.toNumber() +
                openOrdersAccount['referrerRebatesAccrued'].toNumber(), mangoGroup.tokens[MangoGroup_1.QUOTE_INDEX].decimals)));
        }
        return assetsVal;
    }
    getAssetsVal(mangoGroup, mangoCache, healthType) {
        let assetsVal = fixednum_1.ZERO_I80F48;
        // quote currency deposits
        assetsVal = assetsVal.add(this.getUiDeposit(mangoCache.rootBankCache[MangoGroup_1.QUOTE_INDEX], mangoGroup, MangoGroup_1.QUOTE_INDEX));
        for (let i = 0; i < mangoGroup.numOracles; i++) {
            let assetWeight = fixednum_1.ONE_I80F48;
            if (healthType === 'Maint') {
                assetWeight = mangoGroup.spotMarkets[i].maintAssetWeight;
            }
            else if (healthType === 'Init') {
                assetWeight = mangoGroup.spotMarkets[i].initAssetWeight;
            }
            const spotVal = this.getSpotVal(mangoGroup, mangoCache, i, assetWeight);
            assetsVal = assetsVal.add(spotVal);
            // TODO get perp value
        }
        return assetsVal;
    }
    getLiabsVal(mangoGroup, mangoCache, healthType) {
        let liabsVal = fixednum_1.ZERO_I80F48;
        liabsVal = liabsVal.add(this.getUiBorrow(mangoCache.rootBankCache[MangoGroup_1.QUOTE_INDEX], mangoGroup, MangoGroup_1.QUOTE_INDEX));
        for (let i = 0; i < mangoGroup.numOracles; i++) {
            let liabWeight = fixednum_1.ONE_I80F48;
            if (healthType === 'Maint') {
                liabWeight = mangoGroup.spotMarkets[i].maintLiabWeight;
            }
            else if (healthType === 'Init') {
                liabWeight = mangoGroup.spotMarkets[i].initLiabWeight;
            }
            liabsVal = liabsVal.add(this.getUiBorrow(mangoCache.rootBankCache[i], mangoGroup, i).mul(mangoGroup.getPrice(i, mangoCache).mul(liabWeight)));
        }
        return liabsVal;
    }
    getNativeLiabsVal(mangoGroup, mangoCache, healthType) {
        let liabsVal = fixednum_1.ZERO_I80F48;
        liabsVal = liabsVal.add(this.getNativeBorrow(mangoCache.rootBankCache[MangoGroup_1.QUOTE_INDEX], MangoGroup_1.QUOTE_INDEX));
        for (let i = 0; i < mangoGroup.numOracles; i++) {
            const price = mangoCache.priceCache[i].price;
            let liabWeight = fixednum_1.ONE_I80F48;
            if (healthType === 'Maint') {
                liabWeight = mangoGroup.spotMarkets[i].maintLiabWeight;
            }
            else if (healthType === 'Init') {
                liabWeight = mangoGroup.spotMarkets[i].initLiabWeight;
            }
            liabsVal = liabsVal.add(this.getNativeBorrow(mangoCache.rootBankCache[i], i).mul(price.mul(liabWeight)));
        }
        return liabsVal;
    }
    getSpotHealth(mangoCache, marketIndex, assetWeight, liabWeight) {
        const bankCache = mangoCache.rootBankCache[marketIndex];
        const price = mangoCache.priceCache[marketIndex].price;
        const baseNet = this.deposits[marketIndex]
            .mul(bankCache.depositIndex)
            .sub(this.borrows[marketIndex].mul(bankCache.borrowIndex));
        let health = fixednum_1.ZERO_I80F48;
        if (!this.inMarginBasket[marketIndex] ||
            this.spotOpenOrders[marketIndex].equals(utils_1.zeroKey)) {
            if (!baseNet.isNeg()) {
                health = baseNet.mul(assetWeight).mul(price);
            }
            else {
                health = baseNet.mul(liabWeight).mul(price);
            }
        }
        else {
            const openOrders = this.spotOpenOrdersAccounts[marketIndex];
            if (openOrders !== undefined) {
                const quoteFree = fixednum_1.I80F48.fromU64(openOrders.quoteTokenFree.add(openOrders['referrerRebatesAccrued']));
                const quoteLocked = fixednum_1.I80F48.fromU64(openOrders.quoteTokenTotal.sub(openOrders.quoteTokenFree));
                const baseFree = fixednum_1.I80F48.fromU64(openOrders.baseTokenFree);
                const baseLocked = fixednum_1.I80F48.fromU64(openOrders.baseTokenTotal.sub(openOrders.baseTokenFree));
                const bidsBaseNet = baseNet
                    .add(quoteLocked.div(price))
                    .add(baseFree)
                    .add(baseLocked);
                const bidsWeight = !bidsBaseNet.isNeg() ? assetWeight : liabWeight;
                const bidsHealth = bidsBaseNet
                    .mul(bidsWeight)
                    .mul(price)
                    .add(quoteFree);
                const asksBaseNet = baseNet.sub(baseLocked).add(baseFree);
                const asksWeight = !bidsBaseNet.isNeg() ? assetWeight : liabWeight;
                const asksHealth = asksBaseNet
                    .mul(asksWeight)
                    .mul(price)
                    .add(price.mul(baseLocked))
                    .add(quoteFree)
                    .add(quoteLocked);
                health = bidsHealth.min(asksHealth);
            }
        }
        return health;
    }
    getHealth(mangoGroup, mangoCache, healthType) {
        const quoteDeposits = this.getNativeDeposit(mangoCache.rootBankCache[MangoGroup_1.QUOTE_INDEX], MangoGroup_1.QUOTE_INDEX);
        const quoteBorrows = this.getNativeBorrow(mangoCache.rootBankCache[MangoGroup_1.QUOTE_INDEX], MangoGroup_1.QUOTE_INDEX);
        let health = quoteDeposits.sub(quoteBorrows);
        for (let i = 0; i < mangoGroup.numOracles; i++) {
            const spotMarket = mangoGroup.spotMarkets[i];
            const perpMarket = mangoGroup.perpMarkets[i];
            const [spotAssetWeight, spotLiabWeight, perpAssetWeight, perpLiabWeight] = healthType === 'Maint'
                ? [
                    spotMarket.maintAssetWeight,
                    spotMarket.maintLiabWeight,
                    perpMarket.maintAssetWeight,
                    perpMarket.maintLiabWeight,
                ]
                : [
                    spotMarket.initAssetWeight,
                    spotMarket.initLiabWeight,
                    perpMarket.initAssetWeight,
                    perpMarket.initLiabWeight,
                ];
            if (!mangoGroup.spotMarkets[i].isEmpty()) {
                const spotHealth = this.getSpotHealth(mangoCache, i, spotAssetWeight, spotLiabWeight);
                health = health.add(spotHealth);
            }
            if (!mangoGroup.perpMarkets[i].isEmpty()) {
                const perpsCache = mangoCache.perpMarketCache[i];
                health = health.add(this.perpAccounts[i].getHealth(perpMarket, mangoCache.priceCache[i].price, perpAssetWeight, perpLiabWeight, perpsCache.longFunding, perpsCache.shortFunding));
            }
        }
        return health;
    }
    getHealthRatio(mangoGroup, mangoCache, healthType) {
        const health = this.getHealth(mangoGroup, mangoCache, healthType);
        const liabsVal = this.getNativeLiabsVal(mangoGroup, mangoCache, healthType);
        let healthRatio = 100;
        if (liabsVal.gt(fixednum_1.ZERO_I80F48)) {
            healthRatio = (health.toNumber() / liabsVal.toNumber()) * 100;
        }
        return Math.max(Math.min(healthRatio, 100), 0);
    }
    computeValue(mangoGroup, mangoCache) {
        let value = fixednum_1.ZERO_I80F48;
        value = value.add(this.getUiDeposit(mangoCache.rootBankCache[MangoGroup_1.QUOTE_INDEX], mangoGroup, MangoGroup_1.QUOTE_INDEX).sub(this.getUiBorrow(mangoCache.rootBankCache[MangoGroup_1.QUOTE_INDEX], mangoGroup, MangoGroup_1.QUOTE_INDEX)));
        for (let i = 0; i < mangoGroup.numOracles; i++) {
            value = value.add(this.getUiDeposit(mangoCache.rootBankCache[i], mangoGroup, i)
                .sub(this.getUiBorrow(mangoCache.rootBankCache[i], mangoGroup, i))
                .mul(mangoGroup.getPrice(i, mangoCache)));
        }
        // TODO add perp vals
        for (let i = 0; i < this.spotOpenOrdersAccounts.length; i++) {
            const oos = this.spotOpenOrdersAccounts[i];
            if (oos != undefined) {
                value = value.add(fixednum_1.I80F48.fromNumber(utils_1.nativeToUi(oos.baseTokenTotal.toNumber(), mangoGroup.tokens[i].decimals)).mul(mangoGroup.getPrice(i, mangoCache)));
                value = value.add(fixednum_1.I80F48.fromNumber(utils_1.nativeToUi(oos.quoteTokenTotal.toNumber() +
                    oos['referrerRebatesAccrued'].toNumber(), mangoGroup.tokens[MangoGroup_1.QUOTE_INDEX].decimals)));
            }
        }
        return value;
    }
}
exports.default = MangoAccount;
//# sourceMappingURL=MangoAccount.js.map