"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Liquidator = void 0;
/**
This will probably move to its own repo at some point but easier to keep it here for now
 */
const os = __importStar(require("os"));
const fs = __importStar(require("fs"));
const client_1 = require("./client");
const web3_js_1 = require("@solana/web3.js");
const utils_1 = require("./utils");
const ids_json_1 = __importDefault(require("./ids.json"));
const config_1 = require("./config");
const fixednum_1 = require("./fixednum");
const serum_1 = require("@project-serum/serum");
const bn_js_1 = __importDefault(require("bn.js"));
class Liquidator {
    /**
     * Long running program that never exits except on keyboard interrupt
     */
    run() {
        return __awaiter(this, void 0, void 0, function* () {
            const interval = process.env.INTERVAL || 3500;
            const config = new config_1.Config(ids_json_1.default);
            const cluster = (process.env.CLUSTER || 'devnet');
            const groupName = process.env.GROUP || 'mango_test_v3.4';
            const groupIds = config.getGroup(cluster, groupName);
            if (!groupIds) {
                throw new Error(`Group ${groupName} not found`);
            }
            console.log(`Starting liquidator for ${groupName}...`);
            const mangoProgramId = groupIds.mangoProgramId;
            const mangoGroupKey = groupIds.publicKey;
            const payer = new web3_js_1.Account(JSON.parse(process.env.KEYPAIR ||
                fs.readFileSync(os.homedir() + '/.config/solana/devnet.json', 'utf-8')));
            console.log(`Payer: ${payer.publicKey.toBase58()}`);
            const connection = new web3_js_1.Connection(config.cluster_urls[cluster], 'processed');
            const client = new client_1.MangoClient(connection, mangoProgramId);
            const liqorMangoAccountKey = new web3_js_1.PublicKey('6jHCjBA21Tecs1NHEkhqHvqsNoxM6e11vYHppvbpMvV8');
            // eslint-disable-next-line
            while (true) {
                yield utils_1.sleep(interval);
                console.time('groupInfo');
                const mangoGroup = yield client.getMangoGroup(mangoGroupKey);
                const cache = yield mangoGroup.loadCache(connection);
                console.log('calling get all mango accounts');
                const liqorMangoAccount = yield client.getMangoAccount(liqorMangoAccountKey, mangoGroup.dexProgramId);
                const mangoAccounts = yield client.getAllMangoAccounts(mangoGroup);
                const perpMarkets = yield Promise.all(groupIds.perpMarkets.map((perpMarket, index) => {
                    return mangoGroup.loadPerpMarket(connection, index, perpMarket.baseDecimals, perpMarket.quoteDecimals);
                }));
                const spotMarkets = yield Promise.all(groupIds.spotMarkets.map((spotMarket) => {
                    return serum_1.Market.load(connection, spotMarket.publicKey, undefined, groupIds.serumProgramId);
                }));
                const rootBanks = yield mangoGroup.loadRootBanks(connection);
                console.timeEnd('groupInfo');
                console.time('checkAccounts');
                for (let mangoAccount of mangoAccounts) {
                    try {
                        const health = mangoAccount.getHealth(mangoGroup, cache, 'Maint');
                        if (health.lt(fixednum_1.ZERO_I80F48)) {
                            console.log(`Sick account ${mangoAccount.publicKey.toBase58()} health: ${health.toString()}`);
                            yield Promise.all(perpMarkets.map((perpMarket) => {
                                return client.forceCancelPerpOrders(mangoGroup, mangoAccount, perpMarket, payer, new bn_js_1.default(5));
                            }));
                            // Liquidate Spot
                            for (let i = 0; i < mangoGroup.spotMarkets.length - 1; i++) {
                                const spotMarket = spotMarkets[i];
                                const spotMarketInfo = mangoGroup.spotMarkets[i];
                                const spotHealth = mangoAccount.getSpotHealth(cache, i, spotMarketInfo.maintAssetWeight, spotMarketInfo.maintLiabWeight);
                                if (spotHealth.lt(fixednum_1.ZERO_I80F48)) {
                                    const baseRootBank = rootBanks[i];
                                    const quoteRootBank = rootBanks[rootBanks.length - 1];
                                    if (!baseRootBank || !quoteRootBank) {
                                        throw new Error(`Error cancelling spot orders: RootBanks not found for market ${i}`);
                                    }
                                    if (mangoAccount.inMarginBasket[i]) {
                                        yield client.forceCancelSpotOrders(mangoGroup, mangoAccount, spotMarket, baseRootBank, quoteRootBank, payer, new bn_js_1.default(5));
                                    }
                                    let minNet = fixednum_1.ZERO_I80F48;
                                    let minNetIndex = -1;
                                    let maxNet = fixednum_1.ZERO_I80F48;
                                    let maxNetIndex = mangoGroup.tokens.length - 1;
                                    const price = cache.priceCache[i]
                                        ? cache.priceCache[i].price
                                        : fixednum_1.ONE_I80F48;
                                    const netDeposit = mangoAccount
                                        .getNativeDeposit(cache.rootBankCache[i], i)
                                        .sub(mangoAccount.getNativeBorrow(cache.rootBankCache[i], i))
                                        .mul(price);
                                    if (netDeposit.lt(minNet)) {
                                        minNet = netDeposit;
                                        minNetIndex = i;
                                    }
                                    else if (netDeposit.gt(maxNet)) {
                                        maxNet = netDeposit;
                                        maxNetIndex = i;
                                    }
                                    const assetRootBank = rootBanks[maxNetIndex];
                                    const liabRootBank = rootBanks[minNetIndex];
                                    if (assetRootBank && liabRootBank) {
                                        yield client.liquidateTokenAndToken(mangoGroup, mangoAccount, liqorMangoAccount, assetRootBank, liabRootBank, payer, new fixednum_1.I80F48(utils_1.uiToNative(100, mangoGroup.tokens[maxNetIndex].decimals)));
                                    }
                                }
                            }
                            // Liquidate Perps
                            for (let i = 0; i < mangoGroup.perpMarkets.length - 1; i++) {
                                const price = yield mangoGroup.getPrice(i, cache);
                                const perpAccount = mangoAccount.perpAccounts[i];
                                const perpMarketInfo = mangoGroup.perpMarkets[i];
                                const perpMarketCache = cache.perpMarketCache[i];
                                const perpHealth = perpAccount.getHealth(perpMarketInfo, price, perpMarketInfo.maintAssetWeight, perpMarketInfo.maintLiabWeight, perpMarketCache.longFunding, perpMarketCache.shortFunding);
                                if (perpHealth.lt(fixednum_1.ZERO_I80F48)) {
                                    if (perpAccount.basePosition.eq(new bn_js_1.default(0))) {
                                        //await client.liquidateTokenAndPerp(mangoGroup, mangoAccount, liqorMangoAccount, rootBanks[rootBanks.length-1], payer, )
                                    }
                                    else {
                                        yield client.liquidatePerpMarket(mangoGroup, mangoAccount, liqorMangoAccount, perpMarkets[i], payer, new bn_js_1.default(5));
                                    }
                                }
                            }
                        }
                    }
                    catch (err) {
                        console.error(`Failed to process account ${mangoAccount.publicKey.toBase58()}`, err);
                    }
                }
                console.timeEnd('checkAccounts');
            }
        });
    }
}
exports.Liquidator = Liquidator;
new Liquidator().run();
//# sourceMappingURL=liquidator.js.map